@@ -15,110 +15,75 @@ import imp
 import os
 import pkgutil
 import sys
-from itertools import chain
 
 from jedi._compatibility import find_module, unicode
-from jedi import common
 from jedi import debug
-from jedi import cache
-from jedi.parser import fast
+from jedi import settings
+from jedi.common import source_to_unicode, unite
+from jedi.parser.diff import FastParser
 from jedi.parser import tree
+from jedi.parser.utils import save_parser, load_parser, parser_cache
 from jedi.evaluate import sys_path
 from jedi.evaluate import helpers
-from jedi import settings
-from jedi.common import source_to_unicode
 from jedi.evaluate import compiled
 from jedi.evaluate import analysis
-from jedi.evaluate.cache import memoize_default, NO_DEFAULT
+from jedi.evaluate.cache import memoize_default
+from jedi.evaluate.filters import AbstractNameDefinition
 
 
-def completion_names(evaluator, imp, pos):
-    name = imp.name_for_position(pos)
-    module = evaluator.wrap(imp.get_parent_until())
-    if name is None:
-        level = 0
-        for node in imp.children:
-            if node.end_pos <= pos:
-                if node in ('.', '...'):
-                    level += len(node.value)
-        import_path = []
+# This memoization is needed, because otherwise we will infinitely loop on
+# certain imports.
+@memoize_default(default=set())
+def infer_import(context, tree_name, is_goto=False):
+    module_context = context.get_root_context()
+    import_node = tree_name.get_parent_until(tree.Import)
+    import_path = import_node.path_for_name(tree_name)
+    from_import_name = None
+    evaluator = context.evaluator
+    try:
+        from_names = import_node.get_from_names()
+    except AttributeError:
+        # Is an import_name
+        pass
     else:
-        # Completion on an existing name.
+        if len(from_names) + 1 == len(import_path):
+            # We have to fetch the from_names part first and then check
+            # if from_names exists in the modules.
+            from_import_name = import_path[-1]
+            import_path = from_names
 
-        # The import path needs to be reduced by one, because we're completing.
-        import_path = imp.path_for_name(name)[:-1]
-        level = imp.level
+    importer = Importer(evaluator, tuple(import_path),
+                        module_context, import_node.level)
 
-    importer = Importer(evaluator, tuple(import_path), module, level)
-    if isinstance(imp, tree.ImportFrom):
-        c = imp.children
-        only_modules = c[c.index('import')].start_pos >= pos
-    else:
-        only_modules = True
-    return importer.completion_names(evaluator, only_modules)
+    types = importer.follow()
 
+    #if import_node.is_nested() and not self.nested_resolve:
+    #    scopes = [NestedImportModule(module, import_node)]
 
-class ImportWrapper(tree.Base):
-    def __init__(self, evaluator, name):
-        self._evaluator = evaluator
-        self._name = name
-
-        self._import = name.get_parent_until(tree.Import)
-        self.import_path = self._import.path_for_name(name)
-
-    @memoize_default()
-    def follow(self, is_goto=False):
-        if self._evaluator.recursion_detector.push_stmt(self._import):
-            # check recursion
-            return []
-
-        try:
-            module = self._evaluator.wrap(self._import.get_parent_until())
-            import_path = self._import.path_for_name(self._name)
-            from_import_name = None
-            try:
-                from_names = self._import.get_from_names()
-            except AttributeError:
-                # Is an import_name
-                pass
-            else:
-                if len(from_names) + 1 == len(import_path):
-                    # We have to fetch the from_names part first and then check
-                    # if from_names exists in the modules.
-                    from_import_name = import_path[-1]
-                    import_path = from_names
-
-            importer = Importer(self._evaluator, tuple(import_path),
-                                module, self._import.level)
+    if from_import_name is not None:
+        types = unite(
+            t.py__getattribute__(
+                unicode(from_import_name),
+                name_context=context,
+                is_goto=is_goto
+            ) for t in types
+        )
 
+        if not types:
+            path = import_path + [from_import_name]
+            importer = Importer(evaluator, tuple(path),
+                                module_context, import_node.level)
             types = importer.follow()
+            # goto only accepts `Name`
+            if is_goto:
+                types = set(s.name for s in types)
+    else:
+        # goto only accepts `Name`
+        if is_goto:
+            types = set(s.name for s in types)
 
-            #if self._import.is_nested() and not self.nested_resolve:
-            #    scopes = [NestedImportModule(module, self._import)]
-
-            if from_import_name is not None:
-                types = list(chain.from_iterable(
-                    self._evaluator.find_types(t, unicode(from_import_name),
-                                               is_goto=is_goto)
-                    for t in types))
-
-                if not types:
-                    path = import_path + [from_import_name]
-                    importer = Importer(self._evaluator, tuple(path),
-                                        module, self._import.level)
-                    types = importer.follow()
-                    # goto only accepts `Name`
-                    if is_goto:
-                        types = [s.name for s in types]
-            else:
-                # goto only accepts `Name`
-                if is_goto:
-                    types = [s.name for s in types]
-
-            debug.dbg('after import: %s', types)
-        finally:
-            self._evaluator.recursion_detector.pop_stmt()
-        return types
+    debug.dbg('after import: %s', types)
+    return types
 
 
 class NestedImportModule(tree.Module):
@@ -153,10 +118,10 @@ class NestedImportModule(tree.Module):
                                    self._nested_import)
 
 
-def _add_error(evaluator, name, message=None):
+def _add_error(context, name, message=None):
+    # Should be a name, not a string!
     if hasattr(name, 'parent'):
-        # Should be a name, not a string!
-        analysis.add(evaluator, 'import-error', name, message)
+        analysis.add(context, 'import-error', name, message)
 
 
 def get_init_path(directory_path):
@@ -171,8 +136,41 @@ def get_init_path(directory_path):
     return None
 
 
+class ImportName(AbstractNameDefinition):
+    start_pos = (1, 0)
+
+    def __init__(self, parent_context, string_name):
+        self.parent_context = parent_context
+        self.string_name = string_name
+
+    def infer(self):
+        return Importer(
+            self.parent_context.evaluator,
+            [self.string_name],
+            self.parent_context,
+        ).follow()
+
+    def get_root_context(self):
+        # Not sure if this is correct.
+        return self.parent_context.get_root_context()
+
+    @property
+    def api_type(self):
+        return 'module'
+
+
+class SubModuleName(ImportName):
+    def infer(self):
+        return Importer(
+            self.parent_context.evaluator,
+            [self.string_name],
+            self.parent_context,
+            level=1
+        ).follow()
+
+
 class Importer(object):
-    def __init__(self, evaluator, import_path, module, level=0):
+    def __init__(self, evaluator, import_path, module_context, level=0):
         """
         An implementation similar to ``__import__``. Use `follow`
         to actually follow the imports.
@@ -188,36 +186,37 @@ class Importer(object):
         debug.speed('import %s' % (import_path,))
         self._evaluator = evaluator
         self.level = level
-        self.module = module
+        self.module_context = module_context
         try:
-            self.file_path = module.py__file__()
+            self.file_path = module_context.py__file__()
         except AttributeError:
             # Can be None for certain compiled modules like 'builtins'.
             self.file_path = None
 
         if level:
-            base = module.py__package__().split('.')
+            base = module_context.py__package__().split('.')
             if base == ['']:
                 base = []
             if level > len(base):
-                path = module.py__file__()
-                import_path = list(import_path)
-                for i in range(level):
-                    path = os.path.dirname(path)
-                dir_name = os.path.basename(path)
-                # This is not the proper way to do relative imports. However, since
-                # Jedi cannot be sure about the entry point, we just calculate an
-                # absolute path here.
-                if dir_name:
-                    import_path.insert(0, dir_name)
-                else:
-                    _add_error(self._evaluator, import_path[-1])
-                    import_path = []
-                    # TODO add import error.
-                    debug.warning('Attempted relative import beyond top-level package.')
+                path = module_context.py__file__()
+                if path is not None:
+                    import_path = list(import_path)
+                    for i in range(level):
+                        path = os.path.dirname(path)
+                    dir_name = os.path.basename(path)
+                    # This is not the proper way to do relative imports. However, since
+                    # Jedi cannot be sure about the entry point, we just calculate an
+                    # absolute path here.
+                    if dir_name:
+                        import_path.insert(0, dir_name)
+                    else:
+                        _add_error(module_context, import_path[-1])
+                        import_path = []
+                        # TODO add import error.
+                        debug.warning('Attempted relative import beyond top-level package.')
             else:
                 # Here we basically rewrite the level to 0.
-                import_path = tuple(base) + import_path
+                import_path = tuple(base) + tuple(import_path)
         self.import_path = import_path
 
     @property
@@ -225,10 +224,12 @@ class Importer(object):
         """Returns the import path as pure strings instead of `Name`."""
         return tuple(str(name) for name in self.import_path)
 
-    @memoize_default()
     def sys_path_with_modifications(self):
         in_path = []
-        sys_path_mod = list(sys_path.sys_path_with_modifications(self._evaluator, self.module))
+        sys_path_mod = list(sys_path.sys_path_with_modifications(
+            self._evaluator,
+            self.module_context
+        ))
         if self.file_path is not None:
             # If you edit e.g. gunicorn, there will be imports like this:
             # `from gunicorn import something`. But gunicorn is not in the
@@ -245,10 +246,9 @@ class Importer(object):
 
         return in_path + sys_path_mod
 
-    @memoize_default(NO_DEFAULT)
     def follow(self):
         if not self.import_path:
-            return []
+            return set()
         return self._do_import(self.import_path, self.sys_path_with_modifications())
 
     def _do_import(self, import_path, sys_path):
@@ -271,7 +271,7 @@ class Importer(object):
 
         module_name = '.'.join(import_parts)
         try:
-            return [self._evaluator.modules[module_name]]
+            return set([self._evaluator.modules[module_name]])
         except KeyError:
             pass
 
@@ -280,29 +280,27 @@ class Importer(object):
             # the module cache.
             bases = self._do_import(import_path[:-1], sys_path)
             if not bases:
-                return []
+                return set()
             # We can take the first element, because only the os special
             # case yields multiple modules, which is not important for
             # further imports.
-            base = bases[0]
+            parent_module = list(bases)[0]
 
             # This is a huge exception, we follow a nested import
             # ``os.path``, because it's a very important one in Python
             # that is being achieved by messing with ``sys.modules`` in
             # ``os``.
             if [str(i) for i in import_path] == ['os', 'path']:
-                return self._evaluator.find_types(base, 'path')
+                return parent_module.py__getattribute__('path')
 
             try:
-                # It's possible that by giving it always the sys path (and not
-                # the __path__ attribute of the parent, we get wrong results
-                # and nested namespace packages don't work.  But I'm not sure.
-                paths = base.py__path__(sys_path)
+                method = parent_module.py__path__
             except AttributeError:
                 # The module is not a package.
-                _add_error(self._evaluator, import_path[-1])
-                return []
+                _add_error(parent_module, import_path[-1])
+                return set()
             else:
+                paths = method()
                 debug.dbg('search_module %s in paths %s', module_name, paths)
                 for path in paths:
                     # At the moment we are only using one path. So this is
@@ -314,9 +312,10 @@ class Importer(object):
                     except ImportError:
                         module_path = None
                 if module_path is None:
-                    _add_error(self._evaluator, import_path[-1])
-                    return []
+                    _add_error(parent_module, import_path[-1])
+                    return set()
         else:
+            parent_module = None
             try:
                 debug.dbg('search_module %s in %s', import_parts[-1], self.file_path)
                 # Override the sys.path. It works only good that way.
@@ -329,30 +328,41 @@ class Importer(object):
                     sys.path = temp
             except ImportError:
                 # The module is not a package.
-                _add_error(self._evaluator, import_path[-1])
-                return []
+                _add_error(self.module_context, import_path[-1])
+                return set()
 
         source = None
         if is_pkg:
             # In this case, we don't have a file yet. Search for the
             # __init__ file.
-            module_path = get_init_path(module_path)
+            if module_path.endswith(('.zip', '.egg')):
+                source = module_file.loader.get_source(module_name)
+            else:
+                module_path = get_init_path(module_path)
         elif module_file:
             source = module_file.read()
             module_file.close()
 
-        if module_file is None and not module_path.endswith('.py'):
-            module = compiled.load_module(module_path)
+        if module_file is None and not module_path.endswith(('.py', '.zip', '.egg')):
+            module = compiled.load_module(self._evaluator, module_path)
         else:
-            module = _load_module(self._evaluator, module_path, source, sys_path)
+            module = _load_module(self._evaluator, module_path, source, sys_path, parent_module)
+
+        if module is None:
+            # The file might raise an ImportError e.g. and therefore not be
+            # importable.
+            return set()
 
         self._evaluator.modules[module_name] = module
-        return [module]
+        return set([module])
 
-    def _generate_name(self, name):
-        return helpers.FakeName(name, parent=self.module)
+    def _generate_name(self, name, in_module=None):
+        # Create a pseudo import to be able to follow them.
+        if in_module is None:
+            return ImportName(self.module_context, name)
+        return SubModuleName(in_module, name)
 
-    def _get_module_names(self, search_path=None):
+    def _get_module_names(self, search_path=None, in_module=None):
         """
         Get the names of all modules in the search_path. This means file names
         and not names defined in the files.
@@ -360,13 +370,13 @@ class Importer(object):
 
         names = []
         # add builtin module names
-        if search_path is None:
+        if search_path is None and in_module is None:
             names += [self._generate_name(name) for name in sys.builtin_module_names]
 
         if search_path is None:
             search_path = self.sys_path_with_modifications()
         for module_loader, name, is_pkg in pkgutil.iter_modules(search_path):
-            names.append(self._generate_name(name))
+            names.append(self._generate_name(name, in_module=in_module))
         return names
 
     def completion_names(self, evaluator, only_modules=False):
@@ -374,14 +384,14 @@ class Importer(object):
         :param only_modules: Indicates wheter it's possible to import a
             definition that is not defined in a module.
         """
-        from jedi.evaluate import finder
+        from jedi.evaluate.representation import ModuleContext
         names = []
         if self.import_path:
             # flask
             if self.str_import_path == ('flask', 'ext'):
                 # List Flask extensions like ``flask_foo``
                 for mod in self._get_module_names():
-                    modname = str(mod)
+                    modname = mod.string_name
                     if modname.startswith('flask_'):
                         extname = modname[len('flask_'):]
                         names.append(self._generate_name(extname))
@@ -391,15 +401,16 @@ class Importer(object):
                     if os.path.isdir(flaskext):
                         names += self._get_module_names([flaskext])
 
-            for scope in self.follow():
+            for context in self.follow():
                 # Non-modules are not completable.
-                if not scope.type == 'file_input':  # not a module
+                if context.api_type != 'module':  # not a module
                     continue
 
                 # namespace packages
-                if isinstance(scope, tree.Module) and scope.path.endswith('__init__.py'):
-                    paths = scope.py__path__(self.sys_path_with_modifications())
-                    names += self._get_module_names(paths)
+                if isinstance(context, ModuleContext) and \
+                        context.py__file__().endswith('__init__.py'):
+                    paths = context.py__path__()
+                    names += self._get_module_names(paths, in_module=context)
 
                 if only_modules:
                     # In the case of an import like `from x.` we don't need to
@@ -411,12 +422,8 @@ class Importer(object):
 
                     continue
 
-                for names_dict in scope.names_dicts(search_global=False):
-                    _names = list(chain.from_iterable(names_dict.values()))
-                    if not _names:
-                        continue
-                    _names = finder.filter_definition_names(_names, scope)
-                    names += _names
+                for filter in context.get_filters(search_global=False):
+                    names += filter.values()
         else:
             # Empty import path=completion after import
             if not self.level:
@@ -431,25 +438,31 @@ class Importer(object):
         return names
 
 
-def _load_module(evaluator, path=None, source=None, sys_path=None):
+def _load_module(evaluator, path=None, source=None, sys_path=None, parent_module=None):
     def load(source):
         dotted_path = path and compiled.dotted_from_fs_path(path, sys_path)
-        if path is not None and path.endswith('.py') \
-                and not dotted_path in settings.auto_import_modules:
+        if path is not None and path.endswith(('.py', '.zip', '.egg')) \
+                and dotted_path not in settings.auto_import_modules:
             if source is None:
                 with open(path, 'rb') as f:
                     source = f.read()
         else:
-            return compiled.load_module(path)
+            return compiled.load_module(evaluator, path)
         p = path
-        p = fast.FastParser(evaluator.grammar, common.source_to_unicode(source), p)
-        cache.save_parser(path, p)
+        p = FastParser(evaluator.grammar, source_to_unicode(source), p)
+        save_parser(path, p)
         return p.module
 
-    cached = cache.load_parser(path)
-    module = load(source) if cached is None else cached.module
-    module = evaluator.wrap(module)
-    return module
+    if sys_path is None:
+        sys_path = evaluator.sys_path
+
+    cached = load_parser(path)
+    module_node = load(source) if cached is None else cached.module
+    if isinstance(module_node, compiled.CompiledObject):
+        return module_node
+
+    from jedi.evaluate.representation import ModuleContext
+    return ModuleContext(evaluator, module_node)
 
 
 def add_module(evaluator, module_name, module):
@@ -461,18 +474,22 @@ def add_module(evaluator, module_name, module):
         evaluator.modules[module_name] = module
 
 
-def get_modules_containing_name(evaluator, mods, name):
+def get_modules_containing_name(evaluator, modules, name):
     """
     Search a name in the directories of modules.
     """
+    from jedi.evaluate import representation as er
+
     def check_python_file(path):
         try:
-            return cache.parser_cache[path].parser.module
+            parser_cache_item = parser_cache[path]
         except KeyError:
             try:
                 return check_fs(path)
             except IOError:
                 return None
+        else:
+            return er.ModuleContext(evaluator, parser_cache_item.parser.module)
 
     def check_fs(path):
         with open(path, 'rb') as f:
@@ -484,24 +501,34 @@ def get_modules_containing_name(evaluator, mods, name):
                 return module
 
     # skip non python modules
-    mods = set(m for m in mods if not isinstance(m, compiled.CompiledObject))
-    mod_paths = set()
-    for m in mods:
-        mod_paths.add(m.path)
+    used_mod_paths = set()
+    for m in modules:
+        try:
+            path = m.py__file__()
+        except AttributeError:
+            pass
+        else:
+            used_mod_paths.add(path)
         yield m
 
-    if settings.dynamic_params_for_other_modules:
-        paths = set(settings.additional_dynamic_modules)
-        for p in mod_paths:
-            if p is not None:
-                d = os.path.dirname(p)
-                for entry in os.listdir(d):
-                    if entry not in mod_paths:
-                        if entry.endswith('.py'):
-                            paths.add(d + os.path.sep + entry)
+    if not settings.dynamic_params_for_other_modules:
+        return
 
-        for p in sorted(paths):
-            # make testing easier, sort it - same results on every interpreter
-            c = check_python_file(p)
-            if c is not None and c not in mods and not isinstance(c, compiled.CompiledObject):
-                yield c
+    paths = set(settings.additional_dynamic_modules)
+    for p in used_mod_paths:
+        if p is not None:
+            # We need abspath, because the seetings paths might not already
+            # have been converted to absolute paths.
+            d = os.path.dirname(os.path.abspath(p))
+            for file_name in os.listdir(d):
+                path = os.path.join(d, file_name)
+                if path not in used_mod_paths and path not in paths:
+                    if file_name.endswith('.py'):
+                        paths.add(path)
+
+    # Sort here to make issues less random.
+    for p in sorted(paths):
+        # make testing easier, sort it - same results on every interpreter
+        m = check_python_file(p)
+        if m is not None and not isinstance(m, compiled.CompiledObject):
+            yield m
